/*
 * GNOME: The Log Applet
 * (C) 1999 Michael Craig
 *
 * with inspiration and help from
 *  the Webcontrol applet
 * 
 *
 * Author: Michael Craig
 *
 */

#include <config.h>
#include <gnome.h>
#include <glib.h>
#include <applet-widget.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>

GtkWidget *text, *applet, *vscroll, *hbox;

gboolean win_destroyed(GtkWidget *w, gboolean *b);
static void apply_cb(GnomePropertyBox *pb, gint page, gboolean d);
static int read_func(int fd, int bytes, gboolean new_file);
gboolean idle(gpointer d);

static const gchar *log_applet_authors[2] = {
  "Michael Craig <mcraig@orion.olg.com>",
  NULL
};
static gboolean showing_about = FALSE;
static gboolean showing_propz = FALSE;
static gboolean idle_on = FALSE;
const char *privcfgpath_g=NULL, *globcfgpath_g=NULL;

typedef struct {
  gchar *logfile;
  int width;
  int lines;
  int fd_io;
  gboolean fd_updated;
} log_applet_properties;
log_applet_properties propz;
void log_applet_init_properties(log_applet_properties *p)
{
  p->logfile = NULL;
  p->width = 200;
  p->lines = 20;
  p->fd_io = -1;
  p->fd_updated = FALSE;
}

typedef struct {
  GtkWidget *window;
  GtkWidget *file_e, *lines_e, *width_e;
  GtkWidget *table;
} log_applet_property_box;
void log_applet_property_box_build(log_applet_property_box *pb)
{
  pb->window = gnome_property_box_new();
  gtk_window_set_title(GTK_WINDOW(pb->window),
		       _("The Log Applet Properties"));
  gtk_signal_connect(GTK_OBJECT(pb->window), "destroy",
		     GTK_SIGNAL_FUNC(win_destroyed),
		     (gpointer)&showing_propz);

  pb->table = gtk_table_new(3, 3, FALSE);
  gtk_container_set_border_width(GTK_CONTAINER(pb->table), GNOME_PAD);
  gtk_signal_connect(GTK_OBJECT(pb->window), "apply",
		     GTK_SIGNAL_FUNC(apply_cb), (gpointer)pb);
  
  pb->file_e = gtk_entry_new();
  gtk_signal_connect_object(GTK_OBJECT(pb->file_e), "changed",
			    GTK_SIGNAL_FUNC(gnome_property_box_changed),
			    GTK_OBJECT(pb->window));
  pb->width_e = gtk_entry_new_with_max_length(4);
  gtk_signal_connect_object(GTK_OBJECT(pb->width_e), "changed",
			    GTK_SIGNAL_FUNC(gnome_property_box_changed),
			    GTK_OBJECT(pb->window));
  pb->lines_e = gtk_entry_new_with_max_length(4);
  gtk_signal_connect_object(GTK_OBJECT(pb->lines_e), "changed",
			    GTK_SIGNAL_FUNC(gnome_property_box_changed),
			    GTK_OBJECT(pb->window));

  gtk_table_attach(GTK_TABLE(pb->table), gtk_label_new("Filename:"),
		   0, 1, 0, 1,
		   GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND,
		   GNOME_PAD, GNOME_PAD);
  gtk_table_attach(GTK_TABLE(pb->table), pb->file_e,
		   1, 2, 0, 1,
		   GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND,
		   GNOME_PAD, GNOME_PAD);
  gtk_table_attach(GTK_TABLE(pb->table),
		   gtk_label_new("Width of applet:"),
		   0, 1, 1, 2,
		   GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND,
		   GNOME_PAD, GNOME_PAD);
  gtk_table_attach(GTK_TABLE(pb->table), pb->width_e,
		   1, 2, 1, 2,
		   GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND,
		   GNOME_PAD, GNOME_PAD);
  gtk_table_attach(GTK_TABLE(pb->table),
		   gtk_label_new("Lines saved by applet:"),
		   0, 1, 2, 3,
		   GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND,
		   GNOME_PAD, GNOME_PAD);
  gtk_table_attach(GTK_TABLE(pb->table), pb->lines_e,
		   1, 2, 2, 3,
		   GTK_FILL|GTK_EXPAND, GTK_FILL|GTK_EXPAND,
		   GNOME_PAD, GNOME_PAD);
  
  
  gnome_property_box_append_page(GNOME_PROPERTY_BOX(pb->window), pb->table,
				 gtk_label_new(_("Properties")));
}

gboolean win_destroyed(GtkWidget *w, gboolean *b)
{
  if (b) *b = FALSE;
  return FALSE;
}

static void about_cb(AppletWidget *w, gpointer d)
{
  GtkWidget *about;
  
  if (!showing_about) {
    about = gnome_about_new(_("The Log Applet"), "v3*10^(-1)",
			    "(C) 1999",
			    log_applet_authors,
			    _("This GNOME Panel applet waits for data "
			      "to be appended to a file and displays "
			      "it conveniently. It's generally used "
			      "to watch log files."),
			    NULL);
    gtk_signal_connect(GTK_OBJECT(about), "destroy",
		       GTK_SIGNAL_FUNC(win_destroyed),
		       (gpointer)&showing_about);
    gtk_widget_show(about);
    showing_about = TRUE;
  }
}

void start_logging_file()
{
  int fd;

  if (propz.fd_io) close(propz.fd_io);
  propz.fd_updated = TRUE;
  fd = open(propz.logfile, O_RDONLY);
  if (fd<0) {
    perror("open");
    printf("ack! bad file\n"); /* NOTE: this ought to be a dialog */
    exit(1);
  }
  propz.fd_io = fd;
  lseek(propz.fd_io, 0, SEEK_END);
  if (!idle_on) {
    g_idle_add(idle, NULL);
    idle_on = TRUE;
  }
  
  if (privcfgpath_g) {
    gnome_config_push_prefix(privcfgpath_g);
    gnome_config_set_string("/log_applet/propz/logfile", propz.logfile);
    gnome_config_set_int("/log_applet/propz/width", propz.width);
    gnome_config_set_int("/log_applet/propz/lines", propz.lines);
    gnome_config_pop_prefix();
    
    gnome_config_sync();
    gnome_config_drop_all();
  }
}

static void apply_cb(GnomePropertyBox *pbox, gint page, gboolean d)
{
  log_applet_property_box *pb;
  int fd;
  gchar *newfile;

  if (page==0) {
    pb = (log_applet_property_box*)d;
    
    newfile = gtk_entry_get_text(GTK_ENTRY(pb->file_e));
    propz.lines = atoi(gtk_entry_get_text(GTK_ENTRY(pb->lines_e)));
    propz.width = atoi(gtk_entry_get_text(GTK_ENTRY(pb->width_e)));
    
    if (newfile && newfile[0]) {
      if (propz.logfile) g_free(propz.logfile);
      propz.logfile = g_strdup(newfile);
    }
    if (!propz.lines) propz.lines = 20;
    if (!propz.width) {
      propz.width = 200;
    }
    gtk_widget_set_usize(text, propz.width, 48);

    
    start_logging_file();
  }
}

static void properties_cb(AppletWidget *w, gpointer d)
{
  log_applet_property_box *pb;

  if (!showing_propz) {
    pb = g_new(log_applet_property_box, 1);
    log_applet_property_box_build(pb);

    gtk_widget_show_all(pb->window);
    showing_propz = TRUE;
  }
}

static int read_func(int fd, int bytes, gboolean new_file)
{
  char buf[64];
  int r, to_r;
  int total_read=0;
  static int newlines_hit = 0;
  int i, j, point;
  gchar *textbuf;

  if (new_file) {
    newlines_hit = 0;
  }

  for (r=-1; bytes && r; bytes -= r) {
    to_r = (bytes > 64) ? 64 : bytes;
    total_read += r = read(fd, buf, to_r);
    gtk_text_insert(GTK_TEXT(text), NULL, NULL, NULL, buf, r);
    for (i=0; i < r; i++) {
      if (buf[i]=='\n') newlines_hit++;
      if (newlines_hit == propz.lines+1) {
	point = gtk_text_get_point(GTK_TEXT(text));
	gtk_text_set_point(GTK_TEXT(text), 0);
	textbuf = gtk_editable_get_chars(GTK_EDITABLE(text), 0, -1);
	for (j=0; textbuf[j]!='\n'; j++);
	g_free(textbuf);
	gtk_text_freeze(GTK_TEXT(text));
	gtk_text_forward_delete(GTK_TEXT(text), j+1);
	gtk_text_set_point(GTK_TEXT(text), point-j-1);
	newlines_hit--;
	gtk_adjustment_set_value(GTK_ADJUSTMENT(GTK_TEXT(text)->vadj),
			      GTK_ADJUSTMENT(GTK_TEXT(text)->vadj)->upper - 
			      GTK_ADJUSTMENT(GTK_TEXT(text)->vadj)->page_size);
	gtk_text_thaw(GTK_TEXT(text));
      }
    }
  }

  return total_read;
}

gboolean idle(gpointer d)
{
  int fd;
  fd_set set;
  struct stat stats;
  static int size=0;
  gboolean newf=FALSE;

  if (propz.fd_updated) {
    propz.fd_updated = FALSE;
    size=0;
    newf=TRUE;
  }

  if (fstat(propz.fd_io, &stats) < 0) {
    printf("ack! can't get stats.\n"); /* NOTE: this ought to be a dialog */
    exit(1);
  }

  if (!size) {
    size = stats.st_size;
  } else {
    if (stats.st_size > size) {
      size += read_func(propz.fd_io, stats.st_size - size, newf);
    }
  }

  usleep(1000);

  return TRUE;
}

static gint applet_save_session(GtkWidget *w,
				const char *privcfgpath,
				const char *globcfgpath)
{
  /*gnome_config_push_prefix(privcfgpath);
  gnome_config_set_string("/log_applet/propz/logfile", propz.logfile);
  gnome_config_set_int("/log_applet/propz/width", propz.width);
  gnome_config_set_int("/log_applet/propz/lines", propz.lines);
  gnome_config_pop_prefix();

  gnome_config_sync();
  gnome_config_drop_all();*/

  privcfgpath_g = privcfgpath;
  globcfgpath_g = globcfgpath;

  return FALSE;
}

int main(int argc, char **argv)
{
  /* GNOME_Panel_SizeType sizet; *//* see below */
  int fd;

  /* i18n stuff */
  bindtextdomain(PACKAGE, GNOMELOCALEDIR);
  textdomain(PACKAGE);

  /* init applet */
  applet_widget_init("log_applet", VERSION, argc, argv,
		     NULL, 0, NULL);

  applet = applet_widget_new("log_applet");
  if (!applet) {
    g_error("can't create applet!\n");
    exit(1);
  }
  
  log_applet_init_properties(&propz);

  gnome_config_push_prefix(APPLET_WIDGET(applet)->privcfgpath);
  propz.logfile = gnome_config_get_string("/log_applet/propz/logfile");
  propz.width = gnome_config_get_int("/log_applet/propz/width");
  propz.lines = gnome_config_get_int("/log_applet/propz/lines");
  gnome_config_pop_prefix();

  if (!propz.width) propz.width = 200;
  if (!propz.lines) propz.lines = 20;

  if (propz.logfile && propz.logfile[0]) {
    printf("log_applet: beginning with save_sessioned data\n");
    start_logging_file();
  }

  text = gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT(text), FALSE);
  gtk_text_set_word_wrap(GTK_TEXT(text), TRUE);

  /*
  sizet = applet_widget_get_panel_size(APPLET_WIDGET(applet));
  switch (sizet) {
  case SIZE_TINY:
    gtk_widget_set_usize(text, propz.width, 24);
    break;
  case SIZE_STANDARD:
    gtk_widget_set_usize(text, propz.width, 48);
    break;
  case SIZE_LARGE:
    gtk_widget_set_usize(text, propz.width, 64);
    break;
  case SIZE_HUGE:
    gtk_widget_set_usize(text, propz.width, 80);
    break;
  }
  */ /* OOPS: this isn't in gnome-core 1.0.7...is there anything similar? */
  gtk_widget_set_usize(text, propz.width, 48); /* ow. */

  vscroll = gtk_vscrollbar_new(GTK_TEXT(text)->vadj);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), text, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(hbox), vscroll, TRUE, TRUE, 0);

  applet_widget_add(APPLET_WIDGET(applet),
		    hbox);

  applet_widget_register_stock_callback(APPLET_WIDGET(applet),
					"about",
					GNOME_STOCK_MENU_ABOUT,
					_("About..."),
					about_cb, NULL);
  applet_widget_register_stock_callback(APPLET_WIDGET(applet),
					"properties",
					GNOME_STOCK_MENU_PROP,
					_("Properties..."),
					properties_cb,
					NULL);

  gtk_signal_connect(GTK_OBJECT(applet), "save_session",
		     GTK_SIGNAL_FUNC(applet_save_session), NULL);

  gtk_widget_show(text);
  gtk_widget_show(vscroll);
  gtk_widget_show(hbox);
  gtk_widget_show(applet);

  applet_widget_gtk_main();

  return 0;
}





